unit AviWriter_Single_Compressed;

/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//       AviWriter -- a unit to create compressed AVI files                //
//               by Matthias Walter (01.03.2006), with large pieces of     //
//                  code stolen from Elliott Shevin and Anders Melander    //
//       version 1.0. Please send comments, suggestions, and advice        //
//       to matthias.walter@listrik.de                                     //
/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  AviWriter will build an AVI file containing one stream of any          //
//  number of TBitmaps                                                     //
//                                                                         //
//  Properties:                                                            //
//     Bitmaps : A TList containing the Pointer to a TBitmap object        //
//               which become a frame of the AVI video stream. The unit    //
//               allocates and frees the TList, but the caller             //
//               is responsible for managing the TBitmaps                  //
//               Manipulate the list as you would any other TList.         //
//               At least one bitmap is required.                          //
//               Size of first Bitmap saved defines size for whole stream. //
//               Color and B&W images can be mixed                         //
//     FrameTime:                                                          //
//               The duration of each video frame, in milliseconds.        //
//     FileName:                                                           //
//               The name of the AVI file to be created                    //
//                                                                         //
//  Methods:                                                               //
//     AVI_Initiate creates the output file and initiates structures       //
//     AVIAddFrame  adds your frame to the AVI File                        //
//     AVI_Close    closes the AVI file and destroys structures            //
//                                                                         //
//  Programmed with Borland Delphi 2006                                    //
//                                                                         //
//  CODEC:                                                                 //
//     Good result achieved using compression CODECs from                  //
//       - Cinepack                                                        //
//       - Intel YUV                                                       //
//       - DivX 6.1 (You need the DivX Player to view the AVI. This CODEC  //
//                  creates VERY small files of excellent quality          //
//     Works but with poor results, using compression CODECs from          //
//       - Intel R3                                                        //
//       - Indeo 4.5 and 5.1                                               //
//       - Video1                                                          //
//     Other Windows CODEC cause a runtime error                           //
//                                                                         //
//  To Do List:                                                            //
//     - Remember last selected CODEC to avoid selection every time        //
/////////////////////////////////////////////////////////////////////////////

interface

uses
  Windows,Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,
{$ifdef VER90}
  ole2;
{$else}
  ActiveX;
{$endif}

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                      Video for Windows Header                              //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// Adapted from Thomas Schimming's VFW.PAS                                    //
// (c) 1996 Thomas Schimming, schimmin@iee1.et.tu-dresden.de                  //
// (c) 1998,99 Anders Melander                                                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

// TAVIFileInfo dwFlag values
const
 AVIF_HASINDEX		= $00000010;
 AVIF_MUSTUSEINDEX	= $00000020;
 AVIF_ISINTERLEAVED	= $00000100;
 AVIF_WASCAPTUREFILE	= $00010000;
 AVIF_COPYRIGHTED	= $00020000;
 AVIF_KNOWN_FLAGS	= $00030130;

 AVIERR_UNSUPPORTED    = $80044065; // MAKE_AVIERR(101)
 AVIERR_BADFORMAT      = $80044066; // MAKE_AVIERR(102)
 AVIERR_MEMORY         = $80044067; // MAKE_AVIERR(103)
 AVIERR_INTERNAL       = $80044068; // MAKE_AVIERR(104)
 AVIERR_BADFLAGS       = $80044069; // MAKE_AVIERR(105)
 AVIERR_BADPARAM       = $8004406A; // MAKE_AVIERR(106)
 AVIERR_BADSIZE        = $8004406B; // MAKE_AVIERR(107)
 AVIERR_BADHANDLE      = $8004406C; // MAKE_AVIERR(108)
 AVIERR_FILEREAD       = $8004406D; // MAKE_AVIERR(109)
 AVIERR_FILEWRITE      = $8004406E; // MAKE_AVIERR(110)
 AVIERR_FILEOPEN       = $8004406F; // MAKE_AVIERR(111)
 AVIERR_COMPRESSOR     = $80044070; // MAKE_AVIERR(112)
 AVIERR_NOCOMPRESSOR   = $80044071; // MAKE_AVIERR(113)
 AVIERR_READONLY       = $80044072; // MAKE_AVIERR(114)
 AVIERR_NODATA         = $80044073; // MAKE_AVIERR(115)
 AVIERR_BUFFERTOOSMALL = $80044074; // MAKE_AVIERR(116)
 AVIERR_CANTCOMPRESS   = $80044075; // MAKE_AVIERR(117)
 AVIERR_USERABORT      = $800440C6; // MAKE_AVIERR(198)
 AVIERR_ERROR          = $800440C7; // MAKE_AVIERR(199)

 // TAVIStreamInfo dwFlag values
 AVISF_DISABLED	= $00000001;
 AVISF_VIDEO_PALCHANGES = $00010000;
 AVISF_KNOWN_FLAGS = $00010001;

 // defines for uiFlags
 ICMF_CHOOSE_KEYFRAME	= $0001;	// show KeyFrame Every box
 ICMF_CHOOSE_DATARATE	= $0002;	// show DataRate box
 ICMF_CHOOSE_PREVIEW =	$0004;	// allow expanded preview dialog
 ICMF_CHOOSE_ALLCOMPRESSORS	= $0008;	// don't only show those that
						// can handle the input format
						// or input data

 // Palette change data record
 RIFF_PaletteChange: DWORD = 1668293411;
 AVIERR_OK       = 0;
 AVIIF_LIST      = $01;
 AVIIF_TWOCC	   = $02;
 AVIIF_KEYFRAME  = $10;

var
 AVIBitmap: TList;
 AVIBM: Pointer;

type
 TAVIFileInfoW = Record
  dwMaxBytesPerSec,	// max. transfer rate
  dwFlags,		// the ever-present flags
  dwCaps,
  dwStreams,
  dwSuggestedBufferSize,
  dwWidth,
  dwHeight,
  dwScale,
  dwRate,	// dwRate / dwScale == samples/second
  dwLength,
  dwEditCount: DWORD;
  szFileType : array[0..63] of WideChar;		// descriptive string for file type?
 end;
 PAVIFileInfoW = ^TAVIFileInfoW;

 TAVIStreamInfoA = record
  fccType,
  fccHandler,
  dwFlags,                       // Contains AVITF_* flags
  dwCaps                : DWORD;
  wPriority,
  wLanguage             : WORD;
  dwScale,
  dwRate,                        // dwRate / dwScale == samples/second
  dwStart,
  dwLength,                      // In units above...
  dwInitialFrames,
  dwSuggestedBufferSize,
  dwQuality,
  dwSampleSize          : DWORD;
  rcFrame               : TRect;
  dwEditCount,
  dwFormatChangeCount   : DWORD;
  szName                : array[0..63] of AnsiChar;
 end;

 PAVIStream       = Pointer;
 PAVIFile         = Pointer;
 TAVIStreamList   = Array[0..1] of PAVIStream;
 PAVIStreamList   = ^TAVIStreamList;
 APAVISTREAM      = TAVIStreamList;
 APAVICOMPRSTREAM = TAVIStreamList;

 TAVICompressOptions = Packed Record
  fccType,
  fccHandler,
  dwKeyFrameEvery,
  dwQuality,
  dwBytesPerSecond,
  dwFlags		       : DWORD;
  lpFormat		     : Pointer;
  cbFormat		     : DWORD;
  lpParms		       : Pointer;
  cbParms,
  dwInterleaveEvery: DWORD;
 end;
 PAVICompressOptions = ^TAVICompressOptions;

 TAVIStreamInfo = TAVIStreamInfoA;
 PAVIStreamInfo = ^TAVIStreamInfo;
 TAVIPalChange = packed Record
  bFirstEntry,
  bNumEntries: Byte;
  wFlags	   : WORD;
  peNew		   : array[Byte] of TPaletteEntry;
 end;

 PAVIPalChange = ^TAVIPalChange;

Function AVI_Initiate(hwnd: THandle; AVIFileName: String): Boolean;
Function AVIAddFrame(FrameNum: Integer): Boolean;
Function AVI_Close: Boolean;

procedure AVIFileInit; stdcall; external 'avifil32.dll' name 'AVIFileInit';
procedure AVIFileExit; stdcall; external 'avifil32.dll' name 'AVIFileExit';
function AVIFileOpen(var ppfile: PAVIFile; szFile: PChar; uMode: UINT;
                     lpHandler: Pointer): HResult; stdcall;
                     external 'avifil32.dll' name 'AVIFileOpenA';
function AVIFileCreateStream(pfile: PAVIFile; var ppavi: PAVISTREAM;
                             var psi: TAVIStreamInfo): HResult; stdcall;
                             external 'avifil32.dll' name 'AVIFileCreateStreamA';
Function AVIMakeCompressedStream(var ppsCompressed: PAVISTREAM;
     		                         ppsSource        : PAVISTREAM;  //Pointer
		                             var pOptions     : TAVICompressOptions;
 		                             pHandler         : Pointer): HResult;
                                 stdcall; external 'avifil32.dll'
function AVIStreamSetFormat(pavi: PAVIStream; lPos: LongInt; lpFormat: pointer;
                            cbFormat: LongInt): HResult; stdcall;
                            external 'avifil32.dll' name 'AVIStreamSetFormat';
function AVIStreamReadFormat(pavi: PAVIStream; lPos: LongInt; lpFormat: pointer;
                             var cbFormat: LongInt): HResult; stdcall;
                             external 'avifil32.dll' name 'AVIStreamReadFormat';
function AVIStreamWrite(pavi: PAVIStream; lStart, lSamples: LongInt;
                        lpBuffer: pointer; cbBuffer: LongInt; dwFlags: DWORD;
                        var plSampWritten: LongInt;
                        var plBytesWritten: LongInt): HResult;
                        stdcall; external 'avifil32.dll' name 'AVIStreamWrite';
function AVIStreamRelease(pavi: PAVISTREAM): ULONG; stdcall;
                          external 'avifil32.dll' name 'AVIStreamRelease';
function AVIFileRelease(pfile: PAVIFile): ULONG; stdcall;
                        external 'avifil32.dll' name 'AVIFileRelease';
function AVIFileGetStream(pfile: PAVIFile; var ppavi: PAVISTREAM;
                          fccType: DWORD; lParam: LongInt): HResult; stdcall;
                          external 'avifil32.dll' name 'AVIFileGetStream';
function CreateEditableStream(var ppsEditable: PAVISTREAM;
                              psSource: PAVISTREAM): HResult; stdcall;
                              external 'avifil32.dll' name 'CreateEditableStream';
Function AVISaveOptions(hWnd         : THandle;
                        uiFlags      : DWORD;
                        nStreams     : Integer;
                        var ppavi    : APAVIStream;
                        var lpOptions: APAVIStream): HResult;
                        stdcall; external 'avifil32.dll';
Function AVISaveOptionsFree(nStreams: Integer;
                            var ppOptions: Pointer): Hresult;
                            stdcall; external 'avifil32.dll';
Procedure InternalGetDIBSizes(Bitmap            : HBITMAP;
                              var InfoHeaderSize: Integer;
                              var ImageSize     : longInt;
                              PixelFormat       : TPixelFormat);
Function InternalGetDIB(Bitmap          : HBITMAP;
                        Palette         : HPALETTE;
                        var BitmapInfo;
                        var Bits;
                        PixelFormat     : TPixelFormat): Boolean;
Procedure InitializeBitmapInfoHeader(Bitmap     : HBITMAP;
                                     var Info   : TBitmapInfoHeader;
                                     PixelFormat: TPixelFormat);
Function mmioStringToFOURCC(sz: String; uFlags: DWord): HResult; //returns fourcc
                       stdcall; external 'winmm.dll' name 'mmioStringToFOURCCA';

implementation

var
 fFileName     : String;
 pFile         : PAVIFile;
 BitmapInfoSize: Integer;
 EBitmap,
 ExtBitmap     : TBitmap;
 pStream,
 pComprStream  : PAVISTREAM;
 Streams,
 CompOpts      : APAVIStream;
 StreamInfo		 : TAVIStreamInfo;
 opts          : TAVICompressOptions;
 BitmapSize    : Integer;
 BitmapBits	   : Pointer;
 BitmapInfo	   : PBitmapInfoHeader;

Function AVI_Initiate(hwnd: THandle; AVIFileName: String): Boolean;
var
 i, AVIERR: Integer;
 OK       : Boolean;
begin
 Result:= False;

 if AVIBitmap.Count < 1 then
  raise Exception.Create('No Bitmap on the Bitmap List');

 //If anything on the Bitmaps TList is not a bitmap, raise an error
 for i:= 0 to AVIBitmap.Count - 1 do
 begin
  ExtBitmap:= AVIBitmap[i];
  if not(ExtBitmap is TBitmap)
   then raise Exception.Create('Bitmaps[' + IntToStr(i) + '] is not a TBitmap');
 end;

 BitmapInfo:= nil;
 BitmapBits:= nil;

 fFileName:= '';
 DeleteFile(AVIFileName); //delete target file if there
 AVIFileInit;

 ExtBitmap:= AVIBitmap[0];

 try
  //Open AVI file for write operation
  if (AVIFileOpen(pFile, Pchar(AVIFileName),
                 OF_WRITE or OF_CREATE OR OF_SHARE_EXCLUSIVE, nil) <> AVIERR_OK) then
  raise Exception.Create('Failed to create AVI video file: ' + AVIFileName);

  //Allocate the bitmap to which the bitmap on the Bitmaps Tlist will be copied
  EBitmap:= TBitmap.Create;
  EBitmap.PixelFormat:= pf24bit; //Target Pixel Format MUST be 24 Bit !!??
  EBitmap.Height:= ExtBitmap.Height;
  EBitmap.Width:= ExtBitmap.Width;

  //Clear stream header.
  FillChar(StreamInfo, SizeOf(StreamInfo), 0);

  //Set frame rate and scale
  StreamInfo.dwRate:= 25;
  StreamInfo.dwScale:= 1;
  StreamInfo.fccType:= mmioStringToFOURCC('vids', 0);
  SetRect(StreamInfo.rcFrame, 0, 0, ExtBitmap.Width, ExtBitmap.Height);
  pComprStream:= Nil;

  //Open AVI data stream
  if (AVIFileCreateStream(pFile, pStream, StreamInfo) <> AVIERR_OK) then
   raise Exception.Create('Failed to create AVI video stream');

  Streams[0]:= pStream;
  Streams[1]:= Nil;
  CompOpts[0]:= @opts;
  CompOpts[1]:= Nil;

  //User must specify which compressor to use and its options
  AVIerr:= AVISaveOptions(hwnd,
                          ICMF_CHOOSE_KEYFRAME or ICMF_CHOOSE_DATARATE,
                          1,
                          Streams,
                          CompOpts);
  if AVIerr <> 1 then
   raise Exception.Create('User aborted Compression Selection');

  //Make compressed stream
  AVIerr:= AVIMakeCompressedStream(pComprStream,
                                   pStream,
                                   Opts,
                                   Nil);
  if AVIerr <> AVIerr_OK then
   raise Exception.Create('Failed to create compresssed video stream');

  //Determine size of DIB
  InternalGetDIBSizes(EBitmap.Handle, BitmapInfoSize, BitmapSize, EBitmap.PixelFormat);
  if (BitmapInfoSize = 0) then
   raise Exception.Create('Failed to retrieve Bitmap Info');

  //Get DIB header and pixel buffers
  GetMem(BitmapInfo, BitmapInfoSize);
  GetMem(BitmapBits, BitmapSize);

  InternalGetDIB(EBitmap.Handle, 0, BitmapInfo^, BitmapBits^, EBitmap.PixelFormat);

  if (AVIStreamSetFormat(pComprStream, 0, BitmapInfo, BitmapInfoSize) <> AVIERR_OK) then
   raise Exception.Create('Failed to set compressed AVI Stream Format');

  Result:= True;
 except
  //
 end;
end;

Function AVIAddFrame(FrameNum: Integer): Boolean;
var
 Samples_Written,
 Bytes_Written,
 AVIERR          : Integer;
begin
 //Convert original picture format (B&W or color) to 24-Bit Color
 EBitmap.Canvas.Draw(0, 0, ExtBitmap);
 //Copy the Bitmap to a DIB
 InternalGetDIB(EBitmap.Handle, 0, BitmapInfo^, BitmapBits^, EBitmap.PixelFormat);
 //Write Bitmap to stream
 AVIERR:= AVIStreamWrite(pComprStream, FrameNum, 1, BitmapBits, BitmapSize,
                         0, Samples_Written, Bytes_Written);
 if AVIERR <> AVIERR_OK then
  raise Exception.Create('Failed to add frame to AVI. Err=' + IntToHex(AVIERR, 8));

 Result:= True;
end;

Function AVI_Close: Boolean;
begin
 If Assigned(EBitmap) then EBitmap.Free;
 If assigned(pStream) then AviStreamRelease(pStream);
 If assigned(pComprStream) then AviStreamRelease(pComprStream);
 if (BitmapInfo <> nil) then FreeMem(BitmapInfo);
 if (BitmapBits <> nil) then FreeMem(BitmapBits);
 If assigned(pFile) then AviFileRelease(pFile);
 AviFileExit;
 Result:= True;
end;

// --------------
// InternalGetDIB
// --------------
// Converts a bitmap to a DIB of a specified PixelFormat.
//
// Parameters:
// Bitmap	The handle of the source bitmap.
// Pal		The handle of the source palette.
// BitmapInfo	The buffer that will receive the DIB's TBitmapInfo structure.
//		A buffer of sufficient size must have been allocated prior to
//		calling this function.
// Bits		The buffer that will receive the DIB's pixel data.
//		A buffer of sufficient size must have been allocated prior to
//		calling this function.
// PixelFormat	The pixel format of the destination DIB.
//
// Returns:
// True on success, False on failure.
//
// Note: The InternalGetDIBSizes function can be used to calculate the
// nescessary sizes of the BitmapInfo and Bits buffers.
//

Function InternalGetDIB(Bitmap: HBITMAP; Palette: HPALETTE;
                  var BitmapInfo; var Bits; PixelFormat: TPixelFormat): Boolean;
// From graphics.pas, "optimized" for our use
var
 OldPal: HPALETTE;
 DC		 : HDC;
begin
 InitializeBitmapInfoHeader(Bitmap, TBitmapInfoHeader(BitmapInfo), PixelFormat);
 OldPal:= 0;
 DC:= CreateCompatibleDC(0);
 try
  if (Palette <> 0) then
  begin
   OldPal:= SelectPalette(DC, Palette, False);
   RealizePalette(DC);
  end;
  Result:= (GetDIBits(DC, Bitmap, 0, abs(TBitmapInfoHeader(BitmapInfo).biHeight),
            @Bits, TBitmapInfo(BitmapInfo), DIB_RGB_COLORS) <> 0);
 finally
  if (OldPal <> 0) then SelectPalette(DC, OldPal, False);
  DeleteDC(DC);
 end;
end;

// -------------------
// InternalGetDIBSizes
// -------------------
// Calculates the buffer sizes nescessary for convertion of a bitmap to a DIB
// of a specified PixelFormat.
// See the GetDIBSizes API function for more info.
//
// Parameters:
// Bitmap	The handle of the source bitmap.
// InfoHeaderSize
//		The returned size of a buffer that will receive the DIB's
//		TBitmapInfo structure.
// ImageSize	The returned size of a buffer that will receive the DIB's
//		pixel data.
// PixelFormat	The pixel format of the destination DIB.
//

Procedure InternalGetDIBSizes(Bitmap: HBITMAP;
                              var InfoHeaderSize: Integer;
                              var ImageSize: longInt;
                              PixelFormat: TPixelFormat);
//From graphics.pas, "optimized" for our use
var
 Info: TBitmapInfoHeader;
begin
 InitializeBitmapInfoHeader(Bitmap, Info, PixelFormat);
 //Check for palette device format
 if (Info.biBitCount > 8) then
 begin
  //Header but no palette
  InfoHeaderSize:= SizeOf(TBitmapInfoHeader);
  if ((Info.biCompression and BI_BITFIELDS) <> 0) then
   Inc(InfoHeaderSize, 12);
 end
 else //Header and palette
  InfoHeaderSize:= SizeOf(TBitmapInfoHeader) + SizeOf(TRGBQuad) * (1 shl Info.biBitCount);

 ImageSize:= Info.biSizeImage;
end;

// --------------------------
// InitializeBitmapInfoHeader
// --------------------------
// Fills a TBitmapInfoHeader with the values of a bitmap when converted to a
// DIB of a specified PixelFormat.
//
// Parameters:
// Bitmap	The handle of the source bitmap.
// Info		The TBitmapInfoHeader buffer that will receive the values.
// PixelFormat	The pixel format of the destination DIB.
//

{$IFDEF BAD_STACK_ALIGNMENT}
  // Disable optimization to circumvent optimizer bug...
  {$IFOPT O+}
    {$DEFINE O_PLUS}
    {$O-}
  {$ENDIF}
{$ENDIF}

Procedure InitializeBitmapInfoHeader(Bitmap     : HBITMAP;
                                     var Info   : TBitmapInfoHeader;
                                     PixelFormat: TPixelFormat);
//From graphics.pas, "optimized" for our use
var
 DIB	 : TDIBSection;
 Bytes : Integer;

 function AlignBit(Bits, BitsPerPixel, Alignment: Cardinal): Cardinal;
 begin
  Dec(Alignment);
  Result := ((Bits * BitsPerPixel) + Alignment) and not Alignment;
  Result := Result SHR 3;
 end;

begin
 try
  DIB.dsbmih.biSize := 0;
  Bytes:= GetObject(Bitmap, SizeOf(DIB), @DIB);
  if (Bytes = 0) then
   raise Exception.Create('Invalid bitmap');

  if (Bytes >= (sizeof(DIB.dsbm) + sizeof(DIB.dsbmih))) and
     (DIB.dsbmih.biSize >= sizeof(DIB.dsbmih)) then
   Info:= DIB.dsbmih
  else
  begin
   FillChar(Info, sizeof(Info), 0);
   with Info, DIB.dsbm do
   begin
    biSize:= SizeOf(Info);
    biWidth:= bmWidth;
    biHeight:= bmHeight;
   end;
  end;
  Case PixelFormat of
   pf1bit: Info.biBitCount:= 1;
   pf4bit: Info.biBitCount:= 4;
   pf8bit: Info.biBitCount:= 8;
   pf24bit: Info.biBitCount:= 24;
  else
   raise Exception.Create('Invalid pixel foramt');
   // Info.biBitCount := DIB.dsbm.bmBitsPixel * DIB.dsbm.bmPlanes;
  end;
  Info.biPlanes:= 1;
  Info.biCompression:= BI_RGB; //Always return data in uncompressed RGB format
  Info.biSizeImage:= AlignBit(Info.biWidth, Info.biBitCount, 32) *
                     Cardinal(abs(Info.biHeight));
 except
  //
 end;
end;
{$IFDEF O_PLUS}
  {$O+}
  {$UNDEF O_PLUS}
{$ENDIF}

begin
 AVIBitmap:= TList.Create;
end.
